// Copyright (C) 1999-2000 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of ccscript.
// 
// The exception is that, if you link the ccscript library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the ccscript library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name ccscript.  If you copy code from other releases into a copy of
// ccscript, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for ccscript, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include <cc++/config.h>
#include <cc++/macros.h>
#include <cc++/misc.h>
#include <cc++/slog.h>
#include <ctype.h>
#include "script.h"

static int keyindex(const char *keyword)
{
	unsigned key = 0;

	while(*keyword)
		key ^= key << 1 ^ (*(keyword++) & 0x1f);

	return key % KEYWORD_INDEX_SIZE;
}

ScriptCommand::ScriptCommand(const char *cfgfile) :
Keydata(cfgfile), Mutex()
{
	static SCRKEYWORDS interp[] = {
		{"slog", &ScriptInterp::scrSlog, &ScriptCommand::chkHasArgs},
		{"ilog", &ScriptInterp::scrSlog, &ScriptCommand::chkHasArgs},
		{"elog", &ScriptInterp::scrElog, &ScriptCommand::chkHasArgs},
		{"exit", &ScriptInterp::scrExit, &ScriptCommand::chkNoArgs},
		{"set", &ScriptInterp::scrSet, &ScriptCommand::chkHasList},
		{"clear", &ScriptInterp::scrClear, &ScriptCommand::chkHasVars},
		{"goto", &ScriptInterp::scrGoto, &ScriptCommand::chkHasArgs},
		{"call", &ScriptInterp::scrCall, &ScriptCommand::chkHasArgs},
		{"gosub", &ScriptInterp::scrCall, &ScriptCommand::chkHasArgs},
		{"pop", &ScriptInterp::scrPop, &ScriptCommand::chkIgnore},
		{"return", &ScriptInterp::scrReturn, &ScriptCommand::chkNoArgs},
		{"select", &ScriptInterp::scrSelect, &ScriptCommand::chkHasArgs},
		{"size", &ScriptInterp::scrSize, &ScriptCommand::chkHasArgs},
		{"init", &ScriptInterp::scrInit, &ScriptCommand::chkHasList},
		{"if", &ScriptInterp::scrIf, &ScriptCommand::chkHasArgs},
		{"do", &ScriptInterp::scrDo, &ScriptCommand::chkIgnore},
		{"for", &ScriptInterp::scrFor, &ScriptCommand::chkHasList},
		{"loop", &ScriptInterp::scrLoop, &ScriptCommand::chkIgnore},
		{"break", &ScriptInterp::scrBreak, &ScriptCommand::chkIgnore},
		{"continue", &ScriptInterp::scrContinue, &ScriptCommand::chkIgnore},
		{"inc", &ScriptInterp::scrInc, &ScriptCommand::chkHasArgs},
		{"dec", &ScriptInterp::scrDec, &ScriptCommand::chkHasArgs},
		{"once", &ScriptInterp::scrOnce, &ScriptCommand::chkHasArgs},
		{"on", &ScriptInterp::scrOn, &ScriptCommand::chkHasArgs},
		{"pack", &ScriptInterp::scrPack, &ScriptCommand::chkHasList},
		{"unpack", &ScriptInterp::scrUnpack, &ScriptCommand::chkHasList},
		{"packed", &ScriptInterp::scrPacked, &ScriptCommand::chkHasArgs},
		{"trylock", &ScriptInterp::scrTryLock, &ScriptCommand::chkIgnore},
		{"waitlock", &ScriptInterp::scrWaitLock, &ScriptCommand::chkIgnore},
		{"unlock", &ScriptInterp::scrUnlock, &ScriptCommand::chkIgnore},
		{"has", &ScriptInterp::scrHas, &ScriptCommand::chkHasArgs},
		{"missing", &ScriptInterp::scrMissing, &ScriptCommand::chkHasArgs},
		{NULL, NULL, NULL}};

	int i;
	memset(&keywords, 0, sizeof(keywords));
	for(i = 0; i < TRAP_BITS; ++i)
		traps[i] = "<undefined>";

	active = NULL;
	keyword_count = 0;
	trap_count = 0;

	// add special default exit and error traps

	Trap("exit");
	Trap("error");
	Load(interp);
}

scriptmethod_t ScriptCommand::getHandler(const char *keyword)
{
	keyword_t *key = keywords[keyindex(keyword)];

	while(key)
	{
		if(!stricmp(key->keyword, keyword))
			return key->method;

		key = key->next;
	}
	return (scriptmethod_t)NULL;
}

char *ScriptCommand::Check(char *keyword, scriptline_t *line)
{
	keyword_t *key = keywords[keyindex(keyword)];

	while(key)
	{
		if(!stricmp(key->keyword, keyword))
			return Check(key->check, line);

		key = key->next;
	}
	return "unknown command";
}

unsigned ScriptCommand::getTrapId(const char *trapname)
{
	unsigned i;

	for(i = 0; i < TRAP_BITS; ++i)
	{
		if(!stricmp(traps[i], trapname))
			return i;
	}
	return 0;
}

unsigned long ScriptCommand::getTrapMask(unsigned id)
{
	return 1 << id;
}

unsigned long ScriptCommand::getTrapMask(const char *trapname)
{
	unsigned long mask = 1;
	int i;

	for(i = 0; i < TRAP_BITS; ++i)
	{
		if(!stricmp(traps[i], trapname))
			return mask;

		mask = mask << 1;
	}
	return 0;
}

void ScriptCommand::Load(SCRKEYWORDS *keydefs)
{
	int key;
	keyword_t *script;

	for(;;)
	{
		if(!keydefs->keyword)
			break;

		key = keyindex(keydefs->keyword);
		script = (keyword_t *)alloc(sizeof(keyword_t) + strlen(keydefs->keyword));
		strcpy(script->keyword, keydefs->keyword);
		script->method = keydefs->method;
		script->check = keydefs->check;
		script->next = keywords[key];
		keywords[key] = script;
		++keydefs;
	}
}

int ScriptCommand::Trap(const char *trapname)
{
	traps[trap_count++] = alloc((char *)trapname);
	return trap_count;
}

char *ScriptCommand::chkIgnore(scriptline_t *line)
{
	return NULL;
}

char *ScriptCommand::chkHasModify(scriptline_t *line)
{
	if(!line->argc)
		return "arguments missing";

	if(line->argc < 2)
		return "no values to modify";

	if(*(line->args[0]) != '%' && *(line->args[0]) != '@')
		return "invalid variable assignment";

	return NULL;
}

char *ScriptCommand::chkHasList(scriptline_t *line)
{
	if(!line->argc)
		return "arguments missing";

	if(line->argc < 2)
		return "no values to assign";

	if(*(line->args[0]) != '%' && *(line->args[0]) !='@')
		return "invalid variable assignment";

	return NULL;
}

char *ScriptCommand::chkHasArgs(scriptline_t *line)
{
	if(!line->argc)
		return "arguments missing";

	return NULL;
}

char *ScriptCommand::chkNoArgs(scriptline_t *line)
{
	if(line->argc)
		return "invalid arguments";

	return NULL;
} 

char *ScriptCommand::chkHasVars(scriptline_t *line)
{
	int index = 0;
	char ch;

	if(line->argc < 1)
		return "no arguments";

	while(index < line->argc)
	{
		ch = *(line->args[index++]);
		if(ch != '%' && ch != '@')
			return "variable argument required";
	}

	return NULL;
}

ScriptSymbol::ScriptSymbol(int size, int paging) :
SharedMemPager(paging)
{
	symsize = size;
	memset(index, 0, sizeof(index));
}

void ScriptSymbol::Purge(void)
{
	MemPager::purge();
	memset(index, 0, sizeof(index));
}

unsigned ScriptSymbol::getIndex(const char *symname)
{
	unsigned int key = 0;

	while(*symname)
		key ^= (key << 1) ^ (*(symname++) & 0x1f);		
	
	return key % SYMBOL_INDEX_SIZE;
}

scriptsymbol_t *ScriptSymbol::getEntry(const char *symname, int size)
{
	int key = getIndex(symname);
	scriptsymbol_t *sym;

	EnterMutex();
	sym = index[key];
	while(sym)
	{
		if(!stricmp(sym->id, symname))
			break;

		sym = sym->next;
	}
	if(sym || !size)
	{
		LeaveMutex();
		return sym;
	}

	sym = (scriptsymbol_t *)alloc(sizeof(scriptsymbol_t) + size);
	sym->id = MemPager::alloc((char *)symname);
	sym->next = index[key];
	sym->flags.size = size;
	sym->flags.commit = false;
	sym->flags.initial = true;
	sym->flags.system = false;
	sym->flags.readonly = false;
	sym->data[0] = 0;
	index[key] = sym;
	LeaveMutex();
	return sym;
}

char *ScriptSymbol::getSymbol(const char *id)
{
	scriptsymbol_t *sym = getEntry(id);
	if(!sym)
		return NULL;

	if(sym->flags.initial)
		return NULL;

	return sym->data;
}

char *ScriptSymbol::setSymbol(const char *id, const char *value)
{
	scriptsymbol_t *sym = getEntry(id);
	if(!sym)
		return NULL;

	if(sym->flags.readonly)
		return NULL;

	if(!value)
		value = "";

	EnterMutex();
	sym->flags.initial = false;
	strncpy(sym->data, value, sym->flags.size);
	sym->data[sym->flags.size] = 0;

	if(sym->flags.commit)
		Commit(sym);

	LeaveMutex();
	return sym->data;
}

void ScriptSymbol::clrSymbol(const char *id)
{
	scriptsymbol_t *sym = getEntry(id);

	if(!sym)
		return;

	if(sym->flags.readonly || sym->flags.system)
		return;

	EnterMutex();
	sym->flags.initial = true;
	sym->data[0] = 0;
	if(sym->flags.commit)
		Commit(sym);
	LeaveMutex();
	return;
}

char *ScriptSymbol::setSymbol(const char *id, int size)
{
	scriptsymbol_t *sym;

	if(!size)
		size = symsize;

	sym = getEntry(id, size);
	if(!sym)
		return NULL;

	return sym->data;
}

ScriptImage::ScriptImage(ScriptCommand *cmd)
{
	cmds = cmd;
	memset(index, 0, sizeof(index));
	refcount = 0;
}

void ScriptImage::Commit(void)
{
	cmds->EnterMutex();
	cmds->active = this;
	cmds->LeaveMutex();
}

char *ScriptImage::getToken(void)
{
	static char temp[513];
	char *cp = temp;

	if(!quote)
		while(*bp == ' ' || *bp == '\t')
			++bp;

	if(!quote && *bp == '#')
		return NULL;

	if(!*bp)
	{
		quote = false;
		return NULL;
	}

	if(*bp == '\"' && !quote)
	{
		++bp;
		quote = true;
	}

	if(!quote)
	{
		while(*bp && !isspace(*bp) && *bp != ',')
			*(cp++) = *(bp++);

		*cp = 0;
		if(*bp == ',')
			++bp;
		return temp;
	}

	if(isalnum(*bp) || strchr(":., \t\'", *bp))
	{
		while(isalnum(*bp) || strchr("=/:. \t\'", *bp))
			*(cp++) = *(bp++);
	}
	else while(!isspace(*bp) && *bp && *bp != '\"')
		*(cp++) = *(bp++);

	if(*bp == '\n' || !*bp || *bp == '\"')
		quote = false;

	if(*bp == '\n' || *bp == '\"')
		++bp;

	*cp = 0;
	return temp;
}	

scriptname_t *ScriptImage::getScript(const char *name)
{
	int key = keyindex(name);
	scriptname_t *scr = index[key];

	while(scr)
	{
		if(!stricmp(scr->name, name))
			break;

		scr = scr->next;
	}
	return scr;
}

int ScriptImage::Compile(const char *scrname)
{
	char buffer[strlen(scrname) + 1];
	char *token;
	char *ext;

	strcpy(buffer, scrname);
	token = strrchr(buffer, '/');
	if(!token)
		token = buffer;
	else
		++token;

	ext = strrchr(token, '.');
	if(ext)
		*ext = 0;

	return Compile(scrname, token);
}

scriptname_t *ScriptImage::Include(const char *scrname)
{
	char buffer[strlen(scrname) + 1];
	char *token;
	char *ext;
	scriptname_t *inc;

	strcpy(buffer, scrname);
	token = strrchr(buffer, '/');
	if(!token)
		token = buffer;
	else
		++token;

	ext = strrchr(token, '.');
	if(ext)
		*ext = 0;

	inc = getScript(token);
	if(!inc)
	{
		Compile(scrname, token);
		inc = getScript(token);
	}
	return inc;
}


int ScriptImage::Compile(const char *scrname, char *name)
{
	const char *basename = name;
	char namebuf[256];
	char *command, *token;
	char *args[33];
	char *err, *cp;
	bool trapflag;
	int argc, i, key;
	unsigned short count, number, total = 0;
	scriptname_t *script;
	scriptline_t *line, *last;
	unsigned long addmask, submask, trapmask, mask;
	unsigned branch;
	scriptmethod_t handler;
	unsigned char loopid, looplevel;
	scriptname_t *tmpl = NULL, *base = NULL;
	bool sub = false;
	streampos pos;

	scrFile.open(scrname, ios::in);
	if(!scrFile)
		return 0;

compile:
	trapflag = false;
	count = number = 0;
	last = NULL;
	addmask = submask = trapmask = 0;
	branch = 0;
	handler = NULL;
	loopid = looplevel = 0;

	key = keyindex(name);
	script = (scriptname_t *)alloc(sizeof(scriptname_t));	
	memset(script, 0, sizeof(scriptname_t));
	script->name = alloc(name);
	script->mask = 0;
	script->next = index[key];
	index[key] = script;

	if(!base)
		base = script;
 
	if(sub)
	{
		sub = false;
		memcpy(script->trap, base->trap, sizeof(base->trap));
	}
	
	for(;;)
	{	
		quote = false;
		scrFile.getline(buffer, sizeof(buffer) - 1);
		if(scrFile.eof())
			break;

		bp = buffer;
		++number;
		while(NULL != (token = getToken()))
		{
			if(!strncmp(token, "::", 2))
			{
				name = token + 2;
				break;
			}
			if(*token == '^')
			{
				if(!trapflag)
				{
					trapmask = 0;
					trapflag = true;
				}
			}
			if(*token != '^' && *token != '+' && *token != '-')
			{
				if(*token >= '0' && *token <= '9')
				{
					branch |= 1 << (*token - '0');
				}
				else
					break;
			}

			if(*token == '^')
				mask = cmds->getTrapMask(token + 1);
			else
				mask = cmds->getTrapModifier(token + 1);

			if(!mask)
			{
				slog(SLOG_ERROR) << token + 1 << ": unknown trap id; " << script->name << "(" << number << ")" << endl;
				continue;
			}

			switch(*token)
			{
			case '^':
				last = NULL;
				script->mask |= mask | cmds->getTrapDefault();
				trapmask |= mask;
				break;
			case '+':
				addmask |= mask;
				break;
			case '-':
				submask |= mask;
			}
		}
		if(!token)
			continue;

		if(!strncmp(token, "::", 2))
			break;

		if(!stricmp(token, ".module"))
		{
			name = getToken();
			if(name)
				break;
			continue;
		}

		if(!stricmp(token, ".sub"))
		{
			name = getToken();
			if(name)
			{
				sub = true;
				break;
			}
			continue;
		}

		if(!stricmp(token, ".include"))
		{
			token = getToken();
			if(!token)
				continue;
			
			if(*token != '/')
			{
				strcpy(namebuf, scrname);
				cp = strrchr(namebuf, '/');
				if(cp)
				{
					strcpy(cp + 1, token);
					token = namebuf;
				}
			}
			pos = scrFile.tellg();
			scrFile.close();
			Include(token);
			scrFile.open(scrname, ios::in); 
			scrFile.seekg(pos);
			continue;
		}

		if(!stricmp(token, ".template"))
		{
			token = getToken();
			if(*token != '/')
			{
				strcpy(namebuf, scrname);
				cp = strrchr(namebuf, '/');
				if(cp)
				{
					strcpy(cp + 1, token);
					token = namebuf;
				}
			}
			pos = scrFile.tellg();
			scrFile.close();
			tmpl = Include(token);
			scrFile.open(scrname, ios::in);
			scrFile.seekg(pos);
			if(!tmpl)
				continue;

			for(i = 0; i < TRAP_BITS; ++i)
			{
				if(!script->trap[i])
					script->trap[i] = tmpl->trap[i];
			}	
			continue;
		}

		if(!stricmp(token, ".define"))
		{
			while(NULL != (token = getToken()))
			{
				if(!cmds->getLast(token))
					cmds->setValue(token, getToken());
				else
					getToken();
			}
			continue;
		}

		trapflag = false;
		handler = cmds->getHandler(token);
		if(!handler)
		{
			addmask = submask = 0;
			slog(SLOG_ERROR) << token << ": unknown keyword; " << script->name << "(" << number << ")" << endl;
			continue;
		}

		command = alloc(token);	
	
		argc = 0;
		while(argc < 32 && NULL != (token = getToken()))
		{
			if(*token == '$')
			{
				token = (char *)cmds->getLast(++token);
				if(!token)
					token = "";
			}
			else
				token = alloc(token);
			
			args[argc++] = token;
		}

		args[argc++] = NULL;
		line = (scriptline_t *)alloc(sizeof(scriptline_t));
		line->line = number;
		line->mask = ((~0 & ~trapmask) | addmask) & ~submask;
		if(script->mask)
			line->mask &= cmds->getTrapHandler(script);
		line->next = NULL;
		line->args = (char **)alloc(sizeof(char *) * argc);
		line->argc = --argc;
		line->method = handler;
		line->cmd = command;
		line->loop = 0;

		addmask = submask = 0;

		if(!stricmp(command, "for") || !stricmp(command, "do"))
		{
			if(!looplevel)
				++loopid;
			++looplevel;
			line->loop = loopid * 256 + looplevel;
		}

		if(!stricmp(command, "loop"))
		{
			line->loop = loopid * 256 + looplevel;
			if(!looplevel)
			{
				slog(SLOG_ERROR) << "loop: nesting error" << script->name << "(" << number << ")" << endl;
				continue;
			}
			else
				--looplevel;
		}
		memcpy(line->args, &args, sizeof(char *) * argc);

		err = cmds->Check(command, line);
		if(err)
		{
			slog(SLOG_ERROR) << command << ": " << err << "; " << script->name << "(" << number << ")" << endl;
			continue;
		}

		++count;
		script->mask |= trapmask;
		if(!script->first)
			script->first = line;

		if(trapmask && !last)
		{
			for(i = 0; i < TRAP_BITS; ++i)
			{
				if((1l << i) & trapmask)
				{
					if(!script->trap[i])
						script->trap[i] = line;
				}
			}
		}
		if(branch)
		{
			for(i = 0; i < 10; ++i)
			{
				if((1l << i) & branch)
				{
					if(!script->skip[i])
						script->skip[i] = line;
				}
			}
		}

		if(last)
			last->next = line;
			
		last = line;
		branch = 0;
	}
	line = script->first;
	if(!script->mask)
		script->mask = cmds->getTrapDefault();
	while(line)
	{
		line->mask &= script->mask;
		line = line->next;
	}
	total += count;
	slog(SLOG_INFO) << script->name << ": " << count << " steps compiled" << endl;
	if(!scrFile.eof())
	{
		strcpy(namebuf, basename);
		strcat(namebuf, "::");
		strcat(namebuf, name);
		name = namebuf;
		goto compile;
	}
	scrFile.close();
	return total;
}

